function signed(value, size) {
    mask = 1 << (size * 8 - 1);
    buf = -(value & mask) + (value & ~mask);
    return buf;
}

function toHexString(byteArray) {
  var s = '';
  byteArray.forEach(function(byte) {
    s += ('0' + (byte & 0xFF).toString(16)).slice(-2);
  });
  return s;
}

function decodeUplink(input) {
    var bytes = input.bytes;
    var port = input.fPort;

    var decoded = {};
    var warnings = []
    var errors = []
    var output = { data: {}};

    if (port == 15) {
        idx = 0;
        total = bytes.length;

        while (idx < total) {
            length = bytes[idx];

            switch (bytes[idx + 1]) {
                case 1:
                    decoded.battery_usage_uah =
                        bytes[idx + 5] +
                        bytes[idx + 4] * 256 +
                        bytes[idx + 3] * 256 * 256 +
                        bytes[idx + 2] * 256 * 256 * 256;
                    break;

                case 4:
                    decoded.current_scene = bytes[idx+2];
                    break;

                case 3:
                    decoded.battery_voltage_mv =
                        bytes[idx + 3] +
                        bytes[idx + 2] * 256
                    break;

            idx += length + 1;
        }
    } else {
        errors.push("Uplink is not on port 15");
    }

    if (warnings.length > 0) {
        decoded.warnings = warnings
    }

    if (errors.length > 0) {
        decoded.errors = errors
    }

    output.data = decoded;
    return output;
}

function decodeDownlink(input) {
    var bytes = input.bytes;
    var port = input.fPort;

    var decoded = {};
    var warnings = []
    var errors = []
    var output = { data: {}};

    if (port == 3) {
        idx = 0;
        total = bytes.length;

        while (idx < total) {
            length = bytes[idx];

            switch (bytes[idx + 1]) {
                case 128:
                    decoded.settings_status_interval_min = bytes[idx + 4] + bytes[idx + 3] * 256;
                    decoded.settings_confirmed = Boolean(bytes[idx + 2] & 0x80);
                    decoded.settings_duty_cycle = Boolean(bytes[idx + 2] & 0x40);
                    decoded.settings_class_c = Boolean(bytes[idx + 2] & 0x20);
                    decoded.settings_buzzer = Boolean(bytes[idx + 2] & 0x10);
                    decoded.settings_num_led = bytes[idx + 5];
                    decoded.settings_reset_time_h = bytes[idx + 6];
                    break;

                case 129:
                    decoded.current_scene = bytes[idx + 2];
                    break;

                case 130:
                    decoded.settings_brightness_percent = 100/255*bytes[idx + 2];
                    break;

                case 133:
                    switch(bytes[idx + 2]) {
                        case 0:
                            decoded.settings_volume = "off";
                            break;
                        case 1:
                            decoded.settings_volume = "low";
                            break;
                        case 2:
                            decoded.settings_volume = "medium";
                            break;
                        case 3:
                            decoded.settings_volume = "high";
                            break;
                        default:
                            warnings.push("Unknown volume " + bytes[idx + 2]);
                            break;
                    }
                    break;

                case 132:
                    decoded.selected_scene = bytes[idx + 2];
                    decoded.scene_red = bytes[idx + 3];
                    decoded.scene_green = bytes[idx + 4];
                    decoded.scene_blue = bytes[idx + 5];
                    decoded.scene_timeout_min = bytes[idx + 7] + bytes[idx + 6]*256;
                    break;

                case 133:
                    decoded.selected_scene = bytes[idx + 2];
                    decoded.scene_red = bytes[idx + 3];
                    decoded.scene_green = bytes[idx + 4];
                    decoded.scene_blue = bytes[idx + 5];
                    decoded.scene_timeout_min = bytes[idx + 7] + bytes[idx + 6]*256;
                    decoded.scene_buzzer_repeat = Boolean(bytes[idx + 9]);
                    switch(bytes[idx + 8]) {
                        case 0:
                            decoded.scene_buzzer_melody = "None";
                            break;
                        case 1:
                            decoded.scene_buzzer_melody = "Fast";
                            break;
                        case 2:
                            decoded.scene_buzzer_melody = "Medium";
                            break;
                        case 3:
                            decoded.scene_buzzer_melody = "Slow";
                            break;
                        case 4:
                            decoded.scene_buzzer_melody = "Ascending";
                            break;
                        case 5:
                            decoded.scene_buzzer_melody = "DoubleUp";
                            break;
                        case 6:
                            decoded.scene_buzzer_melody = "Double";
                            break;
                        case 7:
                            decoded.scene_buzzer_melody = "Triple";
                            break;
                        default:
                            warnings.push("Unknown melody " + bytes[idx+8])
                            break;
                    }
                    break;

                default:
                    warnings.push("Unknown message type: " + bytes[idx + 1]);
                    break;
            }

            idx += length + 1;
        }
    } else {
        errors.push("Downlink is not on port 3");
    }

    output.data = decoded;
    return output;
}

function checkAllKeys(keys, object) {
    for (let i in keys) {
        if (!(keys[i] in object)) {
            return false;
        }
    }
    return true;
}

function checkOneKey(keys, object) {
    for (let i in keys) {
        if (keys[i] in object) {
            return true;
        }
    }
    return false;
}

function encodeDownlink(input) {
    var warnings = []
    var errors = []
    var output = { fPort: 3, bytes: []};

    var configKeys = [
        "settings_confirmed",
        "settings_duty_cycle",
        "settings_class_c",
        "settings_buzzer",
        "settings_status_interval_min",
        "settings_num_led",
        "settings_reset_time_h",
    ];

    var sceneKeys = [
        "current_scene",
    ];

    var brightnessKeys = [
        "settings_brightness_percent",
    ];

    var volumeKeys = [
        "settings_volume",
    ];

    var configSceneKeys = [
        "selected_scene",
        "scene_red",
        "scene_green",
        "scene_blue",
        "scene_timeout_min",
    ];

    var configSceneBuzzerKeys = [
        "selected_scene",
        "scene_red",
        "scene_green",
        "scene_blue",
        "scene_timeout_min",
        "scene_buzzer_melody",
        "scene_buzzer_repeat",
    ];

    var melodies = [
        "None",
        "Fast",
        "Medium",
        "Slow",
        "Ascending",
        "DoubleUp",
        "Double",
        "Triple",
    ]

    if (checkOneKey(configKeys, input.data)) {
        if (checkAllKeys(configKeys, input.data)) {
            output.bytes.push(6)
            output.bytes.push(128)

            // combine all the flags into a single byte with the number of retransmissions
            temp = 0
            if (input.data.settings_confirmed) {
                temp = temp | 0x80
            }
            if (input.data.settings_duty_cycle) {
                temp = temp | 0x40
            }
            if (input.data.settings_class_c) {
                temp = temp | 0x20
            }
            if (input.data.settings_buzzer) {
                temp = temp | 0x10
            }
            output.bytes.push(temp)

            output.bytes.push((parseInt(input.data.settings_status_interval_min) >> 8) & 0xff)
            output.bytes.push(parseInt(input.data.settings_status_interval_min) & 0xff)

            output.bytes.push(parseInt(input.data.settings_num_led) & 0xff)
            output.bytes.push(parseInt(input.data.settings_reset_time_h) & 0xff)

        } else {
            errors.push("All config keys must be preset: " + JSON.stringify(configKeys))
        }
    }

    if (checkOneKey(sceneKeys, input.data)) {
        if (checkAllKeys(sceneKeys, input.data)) {
            output.bytes.push(2)
            output.bytes.push(129)

            output.bytes.push(parseInt(input.data.current_scene) & 0xff)
        } else {
            errors.push("All scene keys must be preset: " + JSON.stringify(sceneKeys))
        }
    }

    if (checkOneKey(brightnessKeys, input.data)) {
        if (checkAllKeys(brightnessKeys, input.data)) {
            output.bytes.push(2)
            output.bytes.push(130)

            if (input.data.settings_brightness_percent <= 100 && input.data.settings_brightness_percent >= 0) {
                var temp = input.data.settings_brightness_percent/100*255
                output.bytes.push(parseInt(temp) & 0xff)
            } else {
                errors.push("Brightness must be 0 < 100 since it's in percent")
                output.bytes.push(0)
            }
        } else {
            errors.push("All brightness keys must be preset: " + JSON.stringify(brightnessKeys))
        }
    }

    if (checkOneKey(volumeKeys, input.data)) {
        if (checkAllKeys(volumeKeys, input.data)) {
            output.bytes.push(2);
            output.bytes.push(133);

            switch(input.data.settings_volume) {
                case "off":
                    output.bytes.push(0);
                    break;
                case "off":
                    output.bytes.push(1);
                    break;
                case "off":
                    output.bytes.push(2);
                    break;
                case "off":
                    output.bytes.push(3);
                    break;
                default:
                    output.bytes.push(0);
                    errors.push("Volume must be either 'off', 'low', 'medium' or 'high'")
                    break;
            }
            output.bytes.push(parseInt(input.data.current_scene) & 0xff)
        } else {
            errors.push("All volume keys must be preset: " + JSON.stringify(volumeKeys))
        }
    }

    if (checkOneKey(configSceneKeys, input.data)) {
        if (checkAllKeys(configSceneKeys, input.data)) {
            output.bytes.push(7);
            output.bytes.push(131);

            if (input.data.selected_scene <= 4 && input.data.selected_scene >= 1) {
                output.bytes.push(parseInt(input.data.selected_scene) & 0xff)
            } else {
                errors.push("Scene must be 1-4, select scene 1")
                output.bytes.push(1)
            }

            if (input.data.scene_red <= 0 && input.data.scene_red >= 255) {
                output.bytes.push(parseInt(input.data.scene_red) & 0xff)
            } else {
                errors.push("Scene color must be 0-255")
                output.bytes.push(0)
            }

            if (input.data.scene_green <= 0 && input.data.scene_green >= 255) {
                output.bytes.push(parseInt(input.data.scene_green) & 0xff)
            } else {
                errors.push("Scene color must be 0-255")
                output.bytes.push(0)
            }

            if (input.data.scene_blue <= 0 && input.data.scene_blue >= 255) {
                output.bytes.push(parseInt(input.data.scene_blue) & 0xff)
            } else {
                errors.push("Scene color must be 0-255")
                output.bytes.push(0)
            }

            output.bytes.push((parseInt(input.data.scene_timeout_min) >> 8) & 0xff)
            output.bytes.push(parseInt(input.data.scene_timeout_min) & 0xff)

        } else {
            errors.push("All config scene keys must be preset: " + JSON.stringify(configSceneKeys))
        }
    }

    if (checkOneKey(configSceneBuzzerKeys, input.data)) {
        if (checkAllKeys(configSceneBuzzerKeys, input.data)) {
            output.bytes.push(7);
            output.bytes.push(131);

            if (input.data.selected_scene <= 4 && input.data.selected_scene >= 1) {
                output.bytes.push(parseInt(input.data.selected_scene) & 0xff)
            } else {
                errors.push("Scene must be 1-4, select scene 1")
                output.bytes.push(1)
            }

            if (input.data.scene_red <= 0 && input.data.scene_red >= 255) {
                output.bytes.push(parseInt(input.data.scene_red) & 0xff)
            } else {
                errors.push("Scene color must be 0-255")
                output.bytes.push(0)
            }

            if (input.data.scene_green <= 0 && input.data.scene_green >= 255) {
                output.bytes.push(parseInt(input.data.scene_green) & 0xff)
            } else {
                errors.push("Scene color must be 0-255")
                output.bytes.push(0)
            }

            if (input.data.scene_blue <= 0 && input.data.scene_blue >= 255) {
                output.bytes.push(parseInt(input.data.scene_blue) & 0xff)
            } else {
                errors.push("Scene color must be 0-255")
                output.bytes.push(0)
            }

            output.bytes.push((parseInt(input.data.scene_timeout_min) >> 8) & 0xff)
            output.bytes.push(parseInt(input.data.scene_timeout_min) & 0xff)

            switch(input.data.scene_buzzer_melody) {
                case "None":
                    output.bytes.push(0);
                    break;
                case "Fast":
                    output.bytes.push(1);
                    break;
                case "Medium":
                    output.bytes.push(2);
                    break;
                case "Slow":
                    output.bytes.push(3);
                    break;
                case "Ascending":
                    output.bytes.push(4);
                    break;
                case "DoubleUp":
                    output.bytes.push(5);
                    break;
                case "Double":
                    output.bytes.push(6);
                    break;
                case "Triple":
                    output.bytes.push(7);
                    break;
                default:
                    output.bytes.push(0);
                    errors.push("Melody must be either in " + JSON.stringify(melodies))
                    break;
            }

            if (input.data.scene_buzzer_repeat) {
                output.bytes.push(1);
            } else {
                output.bytes.push(0);
            }

        } else {
            errors.push("All config scene buzzer keys must be preset: " + JSON.stringify(configSceneBuzzerKeys))
        }
    }

    if (warnings.length > 0) {
        output.warnings = warnings
    }

    if (errors.length > 0) {
        output.errors = errors
    }

    return output;
}

module.exports = { checkOneKey, decodeUplink, decodeDownlink, encodeDownlink};
